
#include"gaussfgt1D.h"

class fastgb{
 public:
  double scale;
  int IRC;
  fastgb(int dIRC){
    scale=1.0;
    IRC=dIRC;
  }
  virtual ~fastgb(){
  
  }
 inline double gauss(double dx, double dy, double dsig) {
    double dist = dx-dy;
    return exp(-(dist*dist)*dsig);
  }

  
 
//Fast O(N) Methods 
  
  


  


  

  
  void RunFGT_X(Image *in,Image *out,int i,double *intensityX,double sig, gaussfgt1D* mygauss,double error,double *domainX){
    int j;
    
    
    for(j=0;j<in->sx;j++){
      intensityX[j] = in->img[i][j];
    }
    
    mygauss->gaussth2(intensityX,sig,in->sx,domainX,error,IRC);
    
    for(j=0;j<in->sx;j++){
      intensityX[j] = 1.0;
      out->img[i][j] = domainX[j];
      
    }
    mygauss->gaussth2(intensityX,sig,in->sx,domainX,error,IRC);
    for(j=0;j<in->sx;j++){
      double dval = domainX[j];
      if(dval<=0.0)dval=1.0;
      out->img[i][j] /= dval;
    }
  }
  
  

  
  void RunFGT_Y(Image *in,Image *out,int j,double *intensityY,double sig, gaussfgt1D* mygauss,double error,double *domainY){
    int i;
    
    
    for(i=0;i<in->sy;i++){
      intensityY[i] = in->img[i][j];
    }
    
    mygauss->gaussth2(intensityY,sig,in->sy,domainY,error,IRC);
    
    for(i=0;i<in->sy;i++){
      intensityY[i] = 1.0;
      out->img[i][j] = domainY[i];
      
    }
    mygauss->gaussth2(intensityY,sig,in->sy,domainY,error,IRC);
    for(i=0;i<in->sy;i++){
      double dval = domainY[i];
      if(dval<=0.0)dval=1.0;
      out->img[i][j] /= dval;
    }
  }
  


  //FGT1D DomainTransform
  void FastBilateralFilter(Image *iR,Image *iG,Image *iB,double sR,double sG,double sB,double error){
    int i,j;
    double *domainX = new double[iR->sx];
    double *domainY = new double[iR->sy];
    double *intensityX = new double[iR->sx];
    double *intensityY = new double[iR->sy];
    
    Image *tmpR = new Image(iR->sx,iR->sy);
    Image *tmpG = new Image(iR->sx,iR->sy);
    Image *tmpB = new Image(iR->sx,iR->sy);
    gaussfgt1D* mygauss = new gaussfgt1D();
    mygauss->allocate(iR->sx);
    
    
    double vals=(sR+sG+sB)/3.0;
    sR=sG=sB=vals;
    for(i=0;i<iR->sy;i++){
      mygauss->sgx[0]=0.0;
      
      for(j=1;j<iR->sx;j++){
	double gx = iR->img[i][j]-iR->img[i][j-1];
	double gy = iG->img[i][j]-iG->img[i][j-1];
	double gz = iB->img[i][j]-iB->img[i][j-1];
	
	
	mygauss->sgx[j] =mygauss->sgx[(j-1)]+1.0+sqrt((gx*gx+gy*gy+gz*gz));
	
	
      }
      mygauss->scale_to_cube(mygauss->sgx,iR->sx);
      
      RunFGT_X(iR,tmpR,i,intensityX,sR,mygauss,error,domainX);
      RunFGT_X(iG,tmpG,i,intensityX,sG,mygauss,error,domainX);
      RunFGT_X(iB,tmpB,i,intensityX,sB,mygauss,error,domainX);
    }
    
    delete mygauss;
    mygauss = new gaussfgt1D();
    mygauss->allocate(iR->sy);
    for(j=0;j<iR->sx;j++){
      mygauss->sgx[0]=0.0;
      
      for(i=1;i<iR->sy;i++){
	double gx = iR->img[i][j]-iR->img[i-1][j];
	double gy = iG->img[i][j]-iG->img[i-1][j];
	double gz = iB->img[i][j]-iB->img[i-1][j];
	
	
	mygauss->sgx[i] =mygauss->sgx[(i-1)]+1.0+sqrt((gx*gx+gy*gy+gz*gz));
	
	
      }
      mygauss->scale_to_cube(mygauss->sgx,iR->sy);
      
      RunFGT_Y(tmpR,iR,j,intensityY,sR,mygauss,error,domainY);
      RunFGT_Y(tmpG,iG,j,intensityY,sG,mygauss,error,domainY);
      RunFGT_Y(tmpB,iB,j,intensityY,sB,mygauss,error,domainY);
    }
    
    
    delete mygauss;
    
    delete tmpR;
    delete tmpG;
    delete tmpB;
    delete [] domainX;
    delete [] domainY;
    delete [] intensityX;
    delete [] intensityY;
    
    
  }

  //FGT1D DomainTransform Iterative
  void FastBilateralFilter(Image *iR,Image *iG,Image *iB,double sR,double sG,double sB,int N,double error){
    int i,j;
    double *domainX = new double[iR->sx];
    double *domainY = new double[iR->sy];
    double *intensityX = new double[iR->sx];
    double *intensityY = new double[iR->sy];
    
    Image *tmpR = new Image(iR->sx,iR->sy);
    Image *tmpG = new Image(iR->sx,iR->sy);
    Image *tmpB = new Image(iR->sx,iR->sy);
    Image *DomainX = new Image(iR->sx,iR->sy);
    Image *DomainY = new Image(iR->sx,iR->sy);
    
    for(i=0;i<iR->sy;i++){
      DomainX->img[i][0]=0.0;
      for(j=1;j<iR->sx;j++){
	double gx = iR->img[i][j]-iR->img[i][j-1];
	double gy = iG->img[i][j]-iG->img[i][j-1];
	double gz = iB->img[i][j]-iB->img[i][j-1];
	DomainX->img[i][j]=DomainX->img[i][(j-1)]+1.0+sqrt((gx*gx+gy*gy+gz*gz));
      }
    }
    for(j=0;j<iR->sx;j++){
      DomainY->img[0][j]=0.0;
      for(i=1;i<iR->sy;i++){
	double gx = iR->img[i][j]-iR->img[i-1][j];
	double gy = iG->img[i][j]-iG->img[i-1][j];
	double gz = iB->img[i][j]-iB->img[i-1][j];
	DomainY->img[i][j]= DomainY->img[(i-1)][j]+1.0+sqrt((gx*gx+gy*gy+gz*gz));
      }
    }
    

    gaussfgt1D* mygaussX = new gaussfgt1D();
    mygaussX->allocate(iR->sx);
    gaussfgt1D* mygaussY = new gaussfgt1D();
    mygaussY->allocate(iR->sy);
    double vals=(sR+sG+sB)/3.0;
    sR=sG=sB=vals;
    
    
    for(int ite=0;ite<N;ite++){
      for(i=0;i<iR->sy;i++){
	for(j=0;j<iR->sx;j++)
	  mygaussX->sgx[j] =DomainX->img[i][j];
	mygaussX->scale_to_cube(mygaussX->sgx,iR->sx);
	RunFGT_X(iR,tmpR,i,intensityX,sR,mygaussX,error,domainX);
	RunFGT_X(iG,tmpG,i,intensityX,sG,mygaussX,error,domainX);
	RunFGT_X(iB,tmpB,i,intensityX,sB,mygaussX,error,domainX);
      }
      
      
      
      
      for(j=0;j<iR->sx;j++){
	for(i=0;i<iR->sy;i++)	
	  mygaussY->sgx[i] =DomainY->img[i][j];
	
	mygaussY->scale_to_cube(mygaussY->sgx,iR->sy);
	
	RunFGT_Y(tmpR,iR,j,intensityY,sR,mygaussY,error,domainY);
	RunFGT_Y(tmpG,iG,j,intensityY,sG,mygaussY,error,domainY);
	RunFGT_Y(tmpB,iB,j,intensityY,sB,mygaussY,error,domainY);
      }
      sR*=0.5;
      sG*=0.5;
      sB*=0.5;
      
    }
    
    delete mygaussX;
    delete mygaussY;
    delete tmpR;
    delete tmpG;
    delete tmpB;
    delete [] domainX;
    delete [] domainY;
    delete [] intensityX;
    delete [] intensityY;
    delete DomainX;
    delete DomainY;
    
    
  }

  
  void setIDRef(int i,int *di,int dsy){
    
    (*di)=i;
    if(i<0)(*di)=0;
    if(i>=dsy)(*di)=dsy-1;
    
  }
  



};


#undef INF

